/******************************************************************************
*******************************************************************************
**                                                                           **
**      MOST for C Swing and Gopher - A file viewer                          **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: John E. Davis                                  **
**                                                                           **
**      Modifications primarily by Foteos Macrides                           **
**                                                                           **
**      Copyright 1991, John E. Davis                                        **
**                                                                           **
**      Terms of Copyright: See MAIN.C                                       **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
/*
 *  MOST:  buffer.c
 *
 *  Modified for C SWING and GOPHER compatibility by Foteos Macrides
 *                                                MACRIDES@SCI.WFEB.EDU
 */

#include <stdio.h>
#include "buffer.h"
#include "externs.h"
#include "display.h"
#include "line.h"
  
unsigned char *BEG;             /* beginning of current buffer */
unsigned char *EOB;             /* end of current buffer */
unsigned char MINI_BUF[136];

Buffer *BUF;

int NUM_LINES;
int ACTUAL_LINES;

unsigned char *C_POS;
int C_LINE;
unsigned char *M_POS;
int M_FLAG;
int W_LINE;
int TAB_R;
int ESC_NUM;

unsigned char *beg_of_line1()
{
    unsigned char *pos;
    
    if (C_POS == BEG) return BEG;

    pos = C_POS;
    if (pos == EOB) pos--;

    if ((*pos != '\n') && (*(pos-1) == '\n')) return(pos);
    
    if ((*pos == '\n') && (pos == BEG)) return(BEG);
    /* suppose BEG[] is "....\nabcde\n\n\n\n\n..." and we are somwhere in
       the middle of the \n's.  Then we want to return the 2nd \n if the
       SQUEEZE is on otherwise just return where we are. */
    if ((*pos-- == '\n') && (*pos == '\n'))
      {
          if (MOST_S_OPT)
            {
                /* if we are between '\n's then skip past all of them. */
                while ((pos > BEG) && (*pos == '\n')) pos--;
                if ((pos == BEG) && (*pos == '\n')) return (BEG);
                pos++;
            }
          return (++pos);
      }      
    
    while((pos > BEG) && (*pos != '\n')) pos--;
    if (pos != BEG) pos++;
    else if ((pos == BEG) && (*pos == '\n')) pos++;
    return pos;
}

/* does not move point */
unsigned char *end_of_line1()
{
    unsigned char *pos;

    pos = C_POS;
    if (pos >= EOB)  return(EOB);
    /*     if ((pos == BEG) && (*pos == '\n')) return (BEG); */
    
    /* find the first '\n' */
    if (*pos != '\n')
      {
          while ((pos < EOB) && (*pos != '\n')) pos++;
          return(pos);
      }
    
    if (!MOST_S_OPT) return (pos);

    /* if BEG = "....abc\n\n\n\n\ndef..." then we are at some first \n.  We
       want to return the last '\n' unless we wre at the first '\n'. */

    if ((pos > BEG) && ( pos--, *pos++ != '\n')) return (pos); 
    
    while ((pos < EOB) && (*pos == '\n')) pos++;
    if (pos-- == EOB) return (pos);
    if (pos < BEG) pos = BEG;
    return pos;
}

unsigned char *beg_of_line()
{
    unsigned char *b, *i, *e;
    int ii = 0,j,k = 1,esc_num = 0;
    
    b = beg_of_line1();
    
    /*
    **  if not Wrap mode, return beginning of entire line
    */
    if (!MOST_W_OPT) return(b);

    M_FLAG = 0;
    W_LINE = 0;
    TAB_R = 0;
    ESC_NUM = 0;
    e = end_of_line1();
    M_POS = e + 1;

    /*
    **  if a ^M is in the line and not Verbose mode,
    **  set up overstriking for wrapped lines
    */
    if (!MOST_V_OPT)
      for (i = b; i <= e; i++)
        if (*i == '\015')
          {
              M_POS = i;
              M_FLAG = 1;
          }

    /*
    **  wrap mode
    */
    for (i = b; i < C_POS; i++)
      {
          if (i == e - 1)
            {
                return(b);
            }

          /*
          **  make adjustments for non-printing characters
	  **  if not Verbose mode
          */
          else if (!MOST_V_OPT && (*i < ' ' || *i > 126))
            {
                /*
                **  overstrike character via backspace (^H)
                */
                if (*i == '\b' && ii != 0) ii--;

                /*
                **  overstrike line via ^M
                */
                else if (*i == '\015') return(b);
                
                /*
                **  underline, bold, or reverse via <ESC>[#m
                */
                else if (*i == '\033' && i + 3 <= e &&
                          *(i + 1) == '[' &&
                         (*(i + 2) == '1' ||
                          *(i + 2) == '4' ||
                          *(i + 2) == '7' ||
                          *(i + 2) == '0') &&
                          *(i + 3) == 'm')
                  {
                      if      (*(i + 2) == '1') esc_num = 1;
                      else if (*(i + 2) == '4') esc_num = 4;
                      else if (*(i + 2) == '7') esc_num = 7;
                      else esc_num = 0;
                      ii -= 3;
                  }
                else if (*i == '\033' && i + 2 <= e &&
                         *(i + 1) == '[' &&
                         *(i + 2) == 'm')
                  {
                      esc_num = 0;
                      ii -= 2;
                  }

                /*
                **  tab (^I)
                */
                else if (*i == '\t')
                  {
                      j =  8 * (ii/8 + 1) - ii;  /* 8 column tabs */
                      while(j--)
                        {
                            ii++;
                            if (ii == k*SCREEN_WIDTH_M1 + 1 &&
                                *(i + 1) != '\015')
                              {
                                  b = i;
                                  W_LINE++;
                                  TAB_R +=
				    8*(SCREEN_WIDTH_M1/8 + 1) -
                                       SCREEN_WIDTH_M1;
                                  k++;
                                  ESC_NUM = esc_num;
                              }
                        }
                   }

                /*
                **  don't count other control characters except
                **  bell (^G), vertical tab (^K), linefeed (^L),
                **  or escape (^[) not followed by [#m
                */
                else if (*i != '\007' && *i != '\013' &&
                         *i != '\014' && *i != '\033') ;
                else
                  {
                     ii++;
                     if (ii == k*SCREEN_WIDTH_M1 + 1 && *(i + 1) != '\015')
                       {
                          b = i;
                          W_LINE++;
                          TAB_R +=  8*(SCREEN_WIDTH_M1/8 + 1) -
                                    SCREEN_WIDTH_M1;
                          k++;
                          ESC_NUM = esc_num;
                       }
                  }
            }

          /*
          **  printable character or always-explicit control character
          */
          else
            {
                ii++;
                if (ii == k*SCREEN_WIDTH_M1 + 1)
                  {
                      b = i;
                      W_LINE++;
                      TAB_R +=  8*(SCREEN_WIDTH_M1/8 + 1) - SCREEN_WIDTH_M1;
                      k++;
                      ESC_NUM = esc_num;
                  }
            }
      }

    return(b);
}

unsigned char *end_of_line()
{
    unsigned char *b, *e, *i;
    int ii = 0,j;
    
    e = end_of_line1();

    /*
    **  if not Wrap mode, return end of entire line
    */
    if (!MOST_W_OPT) return(e);

    b = beg_of_line();

    /*
    **  wrap mode
    */
    for (i = b; i <= e; i++)
      {
          /*
          **  make adjustments for non-printing characters
	  **  if not Verbose mode
          */
          if (!MOST_V_OPT && (*i < ' ' || *i > 126))
            {
                /*
                **  overstrike character via backspace (^H)
                */
                if (*i == '\b' && ii != 0) ii--;

                /*
                **  overstrike line via ^M
                */
                else if (*i == '\015') return(e);

                /*
                **  underline, bold, or reverse via <ESC>[#m
                */
                else if (*i == '\033' && i + 3 <= e &&
                          *(i + 1) == '[' &&
                         (*(i + 2) == '1' ||
                          *(i + 2) == '4' ||
                          *(i + 2) == '7' ||
                          *(i + 2) == '0') &&
                          *(i + 3) == 'm')
                  {
                      ii -= 3;
                  }
                else if (*i == '\033' && i + 2 <= e &&
                         *(i + 1) == '[' &&
                         *(i + 2) == 'm')
                  {
                      ii -= 2;
                  }

                /*
                **  tab (^I)
                */
                else if (*i == '\t')
                  {
                      if (W_LINE)
                        {
                            if (ii == 0)
                              j = TAB_R;
                            else
                              j = 8 * (ii/8 + 1) - ii + TAB_R;
                        }
                      else
                        j =  8 * (ii/8 + 1) - ii;  /* 8 column tabs */

                      while(j--)
                        {
                            ii++;
                            if (ii == SCREEN_WIDTH + 1)
                              return (i);
                        }
                  }

                /*
                **  don't count other control characters except
                **  bell (^G), vertical tab (^K), linefeed (^L),
                **  or escape (^[) not followed by [#m
                */
                else if (*i != '\007' && *i != '\013' &&
                         *i != '\014' && *i != '\033') ;
                else
                  {
                     ii++;
                     if (ii == SCREEN_WIDTH + 1)
                       {
                          return (i);
                       }
                  }
            }

          /*
          **  printable character or always-explicit control character
          */
          else
            {
                ii++;
                if (ii == SCREEN_WIDTH + 1)
                  {
                      return (i);
                  }
            }
      }

    return (e);
}

int forward_line(int n)
{
    int m, ok;
    int save = n;
    unsigned char *p;
    if (n > 0) 
      {
          if (MOST_B_OPT)
            {
                m = (EOB - C_POS)/16;
                if (n > m) n = m;
                C_POS += n * 16;
                C_LINE += n;
                return n;
            }
          else while (n--)
            {
                C_POS = end_of_line();
                /* next step handles newline at the EOB */
                if ((C_POS == EOB) || ((C_POS == EOB - 1) && (*C_POS == '\n'))) 
                  return (save - n - 1);
                C_LINE++;
                C_POS++;
                if (SELECTIVE_I_DISPLAY && !MOST_W_OPT)
                  {
                      p = C_POS;
                      ok = 1;
                      while ((*p <= ' ') && ok)
                        {
                            if (*p != '\n') p++;
                            if (p >= EOB) break;
                            if ((*p == '\n') ||
                                (apparant_distance(p) >= SELECTIVE_I_DISPLAY))
                              {
                                  ok = 0;
                                  C_LINE--;
                                  n++;
                              }
                        }
                  } /* SELECTIVE_I_DISPLAY */
            }
      }
    else
      {
          if (MOST_B_OPT)
            {
                m = (BEG - C_POS)/16;
                if (n < m) n = m;
                C_POS += n * 16;
                C_LINE += n;
                return n;
            }
          else while (n++)
            {
                C_POS = beg_of_line();
                if (C_POS == BEG) return (n - save - 1);
                C_POS--;
                C_LINE--;
                if (SELECTIVE_I_DISPLAY && !MOST_W_OPT)
                  {
                      C_POS = beg_of_line();
                      p = C_POS;
                      ok = 1;
                      while ((*p <= ' ') && ok)
                        {
                            if (*p != '\n') p++;
                            if (p >= EOB) break;
                            if ((*p == '\n') ||
                                (apparant_distance(p) >= SELECTIVE_I_DISPLAY))
                              {
                                  ok = 0;
                                  C_LINE++;
                                  n--;
                              }
                        }
                  } /* SELECTIVE_I_DISPLAY */
            }
      }
    return(save);
}

int count_lines(unsigned char *beg, unsigned char *end)
{
    int save_line = C_LINE, n;
    unsigned char *save_beg = beg, *save_eob = EOB, *save_pos = C_POS;

    if (MOST_B_OPT)
      {
          n = (int)(end - beg)/16;
          if ((n * 16) == (end - beg)) return(n);
          else return(n + 1);
      }
    BEG = C_POS = beg;
    EOB = end;
    n = 1;
    while(forward_line(1)) n++;
    if (*end ==  '\n') n--;
    if (!n) n = 1;
    C_POS = save_pos;
    EOB = save_eob;
    BEG = save_beg;
    C_LINE = save_line;
    return(n);
}

void goto_line(int line)
{
    register int i,j;
    unsigned char *pos;
    int dif_c, dif_b,dif_t;

    if (line < 1) line = 1;
    if (line > NUM_LINES) line = NUM_LINES;

    if (MOST_B_OPT)
      {
          C_POS = BEG + (16 * (line - 1));
          C_LINE = line;
          return;
      }
    
    dif_c = line - C_LINE;
    dif_b = line - NUM_LINES;
    dif_t = line - 1;

    /* 4 possibilites */
    if (dif_c <= 0)
      {
          if (dif_t < -dif_c) /* go from top */
            {
                C_LINE = 1;
                C_POS = BEG;
                (void) forward_line(dif_t);
            }
          else  /* from curr back */
            {
                (void) forward_line(dif_c);
            }
      }
    else if (dif_c > 0)
      {
          if ((dif_c + dif_b) < 0) /* go from curr */
            {
                (void) forward_line(dif_c);
            }
          else
            {
                C_LINE = NUM_LINES;
                C_POS = EOB;
                (void) forward_line(dif_b);
            }
      }
}       

/* return line the point is on without the final '\n's
    unless beg = end in which case we take care of it later ...

    returns 1 if the line should be wrapped */
int extract_line(unsigned char **beg, unsigned char **end)
{
    *beg = beg_of_line();
    *end = end_of_line();
    if (*end != end_of_line1()) return(1);
    while ((*end > BEG) && (**end == '\n')) *end = *end - 1;
    return(0);
}    

int what_line(unsigned char *pos)
{
    
    unsigned char *save_pos;
    int save_line, dir;
    register int dif_c, dif_b,dif_t;
    int ret;
    extern int S_DIFF;

    if (MOST_B_OPT)
      {
          return (1 + (pos - BEG)/16);
      }
    
    save_pos = C_POS;
    save_line = C_LINE;
    
    dif_c = pos - C_POS;
    dif_b = pos - EOB;
    dif_t = pos - BEG;

    /* 4 possibilites */
    if (dif_c <= 0)
      {
          if (dif_t < -dif_c) /* go from top */
            {
                C_LINE = 1;
                C_POS = BEG;
                dir = 1;
            }
          else  /* from curr back */
            {
                dir = -1;
            }
      }
    else if (dif_c > 0)
      {
          if ((dif_c + dif_b) < 0) /* go from curr */
            {
                dir = 1;
            }
          else
            {
                C_LINE = NUM_LINES;
                C_POS = EOB;
                dir = -1;
            }
      }
    if (dir == 1)
      {
          while(C_POS = end_of_line(), C_POS < pos)
            {
                C_POS++;
                C_LINE++;
            }

          /*
          **  if this is the unseen portion of a wrapped line,
          **  the display puts it on the next line
          */
          if (MOST_W_OPT &&
              (C_POS == pos ||
               (C_POS != end_of_line1() &&
                apparant_distance(pos) + S_DIFF >= SCREEN_WIDTH)))
            C_LINE++;
      }
    else
      {
          while(C_POS = beg_of_line(), pos < C_POS)
            {
                C_LINE--;
                C_POS--;
            }
      }

    ret = C_LINE;
    C_POS = save_pos;
    C_LINE = save_line;
    return(ret);
}

/*
**  Given a buffer position, find the line and column.
*/
void find_row_column(unsigned char *pos, int *r, int *c)
{
    unsigned char *beg, *save_pos;
    int save_line,p;
    extern int S_DIFF;

    if (pos <= BEG)
      {
          *r = 1;
          *c = 1;
          return;
      }

    save_line = C_LINE;
    save_pos = C_POS;
    *r = what_line(pos);
    
    /* 
    **  Now we have found the line it is on so get the column.
    */
    if (MOST_B_OPT)
      {
          *c = (unsigned char *)((int)(pos) - (((int)pos/16) * 16) + 41);
          return;
      }

    C_LINE = *r;
    C_POS = pos;
    *c = apparant_distance(pos) + S_DIFF;
    if (MOST_W_OPT)
      if (end_of_line() != end_of_line1())
        if (*c >= SCREEN_WIDTH) *c -= (SCREEN_WIDTH - 1);

    C_LINE = save_line;
    C_POS = save_pos;
}       

Buffer *switch_to_buffer(Buffer *new)
{
    Buffer *old;
    old = BUF;
    BUF = new;
    BEG = BUF->beg;
    EOB = BUF->end;
    return(old);
}

void delete_buffer(Buffer *old)
{
    (void) free(BUF);
    BUF = old;
    BEG = BUF->beg;
    EOB = BUF->end;
}

Buffer *create_buffer(char *file)
{
    Buffer *buf;
    char *MOST_PROMPT;

    buf = (Buffer *) malloc(sizeof(Buffer));
    buf->mark = 0;
    if (MOST_H_OPT)
        strcpy(buf->file,
	       "MOST.DOC_FM   Press H for command summary, Q to quit.");
    else if (MOST_N_OPT)
      {
        if ((MOST_PROMPT = getenv("MOST_PROMPT")) == NULL)
             MOST_PROMPT = "MOST - Press H for Help, Q to Quit";
        strcpy(buf->file,MOST_PROMPT);
      }
    else
        strcpy(buf->file,file);
    return(buf);
}
