/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_5
*/
#include <stdio.h>
#include <smgdef.h>
#include <ssdef.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

delete_node(del_ptr, del_level)
LINE *del_ptr;
int del_level;
{
    NODE *dn,*ds;

    dn = del_ptr->n_ptr[del_level];
    ds = dn->sister;
    /*
    **  remove from structure and replace with sister
    */

    if (del_ptr->n_ptr[del_level-1])
    {
        /* 
        **  dn is a child - has a parent
        */
        del_ptr->n_ptr[del_level-1]->child = ds;
    }
    else
    {
        /*
        **  dn is somebody's sister
        **  link my older sister to my younger sister
        */
                
        do
        {
            /*
            **  back up a line looking for older sister
            */
            del_ptr = del_ptr->back_ptr;
        } while(del_ptr->n_ptr[del_level] == 0);

        del_ptr->n_ptr[del_level]->sister = ds;
    }
    dn->sister = 0; 
}

add_node(add_ptr, add_level, node)
LINE *add_ptr;
int add_level;
NODE *node;
{
    NODE *pn;

    pn = add_ptr->n_ptr[add_level];

    if (!pn->child)
    {
        /*
        **  if we have no child then the new boy becomes our child
        */
        pn->child = node;
    }
    else
    if (strcmp(pn->child->name,node->name) > 0)
    {
        node->sister = pn->child;
        pn->child = node;
    }
    else
    {
        /*
        **  tie him into our sisters, hopefully preserving
        **  alphabetical order
        */
        pn = pn->child;
        while(pn->sister && (strcmp(pn->sister->name,node->name) < 0))
        {
            pn = pn->sister;
        }
        node->sister = pn->sister;
        pn->sister = node;
    }
}

move_node(new_ptr, new_level, old_ptr, old_level)
LINE *new_ptr, *old_ptr;
int new_level, old_level;
{
    NODE *node;

    node = old_ptr->n_ptr[old_level];

    delete_node(old_ptr, old_level);
    add_node(new_ptr, new_level, node);
}

delete_dirs(option)
int option;
{
     if (!current_level)
     {
         warning("Can't delete main directory");
         return(FALSE);
     } 
     
     show_deleting(curr_l_ptr->n_ptr[current_level],
                   DELETING_RENDITION,FALSE);

     if (!confirm_user("Delete these directories Y/N [N] ?","Y") ||
         !confirm_user("LAST CHANCE - Really delete these directories Y/N [N] ?",
                       "Y"))
     {
         show_deleting(curr_l_ptr->n_ptr[current_level],NORMAL_RENDITION);
         render(current_line, current_level,SPECIAL_RENDITION);
         return(FALSE);
     }
     else
     {
         NODE *node;
         int line,level,i,status,channel;
         char dir[SPEC_SIZE+1];


         sprintf(dir,"[-]%s%s",curr_l_ptr->n_ptr[current_level]->name,
                     ".dir;1");

         /*
         **  now try to delete this directory and all sub-dirs and files
         **  therein
         */
         delete_many_files(curr_l_ptr->n_ptr[current_level],TRUE,DELETE,"*.*;*");

         curr_l_ptr->dp->to_be_saved = TRUE;
         node = curr_l_ptr->n_ptr[current_level];

         /*
         **  now return node storage used by all the children of this node
         */
         free_nodes(node->child);

         /*
         **  make a note of where we are for possible later use
         */
 	 line = current_line;
         level = current_level;
   
         if (access(dir,0) == 0)
         {
             /*
             **  dir still exists
             **
             **  search out undeleted dirs owned by this dir, linking them
             **  in to the node lists
             */
             warning("some files were not deleted - searching ...");
             node->child = 0;

             /*
             **  assign channel to current device
             */
             status = sys$assign(DSC1("SYS$DISK"),&channel,0,0);
             if (status != SS$_NORMAL)
                 panic(status,"sys$asign","search nodes");

             get_dirs(node,channel);

             status = sys$dassgn(channel);
             if (status != SS$_NORMAL)
                 panic(status,"sys$dassgn","search nodes");

             message("some files were not deleted ...");
         }
         else
         {
             /*
             **  dir is not there
             */
             message("Directories and contents deleted");

             /*
             **  unlink node from the node structure
             */
             delete_node(curr_l_ptr,current_level);

             /*
             **  now determine current field - parent of the node we just 
    	     **  deleted
             */
             current_level--;
             while(!curr_l_ptr->n_ptr[current_level])
             {
                 curr_l_ptr = curr_l_ptr->back_ptr;
                 current_line--;
             }

             /*
             **  free storage used by this node
             */
             remove_visited_dir(node);
             free(node);

             /*
             **  record where we are
             */
             line = current_line;
             level = current_level;
         }

         /*
         **  now rebuild the lines and the display, 99% chance that it's 
    	 **  changed, and even if not we want to be sure that we show what
         **  we believe to be true
         */
         load_lines();

         current_line = line;
         current_level = level;

         curr_l_ptr = first_line_ptr;
         for (i = 0; i < line; i++)
             curr_l_ptr = curr_l_ptr->forw_ptr;

         /*
         **  refresh output and change def dir accordingly
         */
         repair_display();
    	 set_default(curr_l_ptr->n_ptr[current_level]->dir_name);
     }
}

add_dirs()
{
    if (current_level < MAX_ADD_LEVEL)
    {
        NODE *node, *this_node;
        int i, status;
        int line, level;
        short fid[3];
        char name[NAME_SIZE+1];
        char dir[SPEC_SIZE+1];

        if (get_user_input("New directory name ? [NONE] ",name,NAME_SIZE))
        {
            if (strpbrk(name,".") != NULL)
            {
                warning("You can only create one directory at a time");
                return(FALSE);
            }

            /*
            **  point to current node
            */
            this_node = curr_l_ptr->n_ptr[current_level];

            /*
            **  build up the new dir string in form 
            **  [existing.new_dir] ready for lib$create_dir
            */
            strcpy(dir,this_node->dir_name);
            sprintf(&dir[strlen(dir)-1],"%c%s%c",'.',name,']');

            /*
            **  if we can't add it then the new name is rotten
            */
            status = lib$create_dir(DSC1(dir));
            if (status != SS$_NORMAL && status != SS$_CREATED)
            {
                sprintf(junk,"%s - can't create directory %s%s%s",
                             full_error_text(status),"[.",name,"]");
                warning(junk);
                return(FALSE);
            }

            /*
            **  directory already there
            */
            if (status == SS$_NORMAL)
            {
                sprintf(junk,"directory %s%s%s already exists",
                             "[.",name,"]");
                warning(junk);
                return(FALSE);
            }

            /*
            **  we created it
            */
            sprintf(junk,"Created %s%s%s","[.",name,"]");
            message(junk);
            curr_l_ptr->dp->to_be_saved = TRUE;

            /*
            ** build up node, hand build dir name and fid
            */
            node = grab_node(name);
    
            /*
            **  directory specification
            */
            strcpy(node->dir_name,dir);

            /*
            **  get fid of new directory
            */
            get_did(dir,node->fid);

            /*
            **  load up parent fid
            */
            node->did[0] = this_node->fid[0];
            node->did[1] = this_node->fid[1];
            node->did[2] = this_node->fid[2];

            add_node(curr_l_ptr, current_level, node);

            /*
            **  rebuild lines
            */
            line = current_line;
            level = current_level;
            load_lines();
            repair_display();
            current_line = line;
            current_level = level;
            curr_l_ptr = first_line_ptr;
            for (i = 0; i < line; i++)
                curr_l_ptr = curr_l_ptr->forw_ptr;
	}
    } 
    else
        warning("Not possible");
}

rename_dir()
{
    if (current_level)
    {
        NODE *node;
        int i;
        char name[NAME_SIZE+1];
        char old_name[SPEC_SIZE+1];
        char new_name[SPEC_SIZE+1];
        char ren_name[SPEC_SIZE+1];

        if (get_user_input("New directory name ? [NONE] ",name,NAME_SIZE))
        {
            /*
            **  now try the rename via lib_rename
            */
            node = curr_l_ptr->n_ptr[current_level];
            sprintf(old_name,"[-]%s%s",node->name,".dir;1");
            sprintf(new_name,"[-]%s%s",name,".dir;1");

#ifdef debug
            sprintf(junk,"lib_rename(%s,%s)",old_name,new_name);
            message(junk);
#endif      
            i = lib_rename(old_name,new_name,ren_name);
            if (i != SS$_NORMAL)
            {
                sprintf(junk,"%s - failed to rename %s%s%s",
                        full_error_text(i),"[.",node->name,"]");
                warning(junk);
                return(FALSE);
            }

            curr_l_ptr->dp->to_be_saved = TRUE;
            /*
            **  find name from ren_name - strip off .type;version
            */
            for (i = 0; i < strlen(ren_name); i++)
                if (ren_name[i] == '.')
                {
                    ren_name[i] = 0;
                    break;
                }
            strcpy(name,ren_name);
            give_node_a_name(node,name);
            delete_node(curr_l_ptr, current_level);
            /*
            **  now travel to parent
            */
            current_level--;
            while(!curr_l_ptr->n_ptr[current_level])
            {
                /*
                **  go back a line
                */
                curr_l_ptr = curr_l_ptr->back_ptr;
                current_line--;
            }
            add_node(curr_l_ptr, current_level, node);

            /*
            **  rebuild lines
            */
            load_lines();
 
            /*
            **  now find new line, level etc. for renamed node
            */
            curr_l_ptr = first_line_ptr;
            for (current_line = 0; current_line < total_lines; current_line++)
            {
                for (current_level = 0; 
                     current_level < MAX_LEVEL; 
                     current_level++)
                {
                    if (curr_l_ptr->n_ptr[current_level] == node)
                    {
                        return(repair_display());
                    }
                }
                curr_l_ptr = curr_l_ptr->forw_ptr;
            }
	}
    } 
    else
        warning("Cannot rename main directory");
}

move_dir()
{
    if (current_level)
    {
        int i,j;
        char name[NAME_SIZE+1];
        NODE *init_node, *parent_node;
        LINE *init_line_ptr, *temp;
        int init_line, init_level;
        char old_name[SPEC_SIZE+1];
        char new_name[SPEC_SIZE+1];

        init_line_ptr = curr_l_ptr;
        init_line     = current_line;
        init_level    = current_level;
        init_node     = curr_l_ptr->n_ptr[current_level];

        travel_to_dir();
        message_set = 0;

        if (next_command(FALSE))
        {
            /*
            **  user selected a new parent node - oh joy oh wonder
            */
            parent_node = curr_l_ptr->n_ptr[current_level];

	    /*
	    **  make sure they don't enter a directory into itself!
	    */
	    if (init_node->dir_name == parent_node->dir_name)
	    {
		warning("Moving a directory under itself is not allowed!");
		message("Directory move aborted.");
		return(FALSE);
	    }
            /*
            **  find init_node full spec
            */
            strcpy(old_name,init_node->dir_name);
            for (i = strlen(old_name) - 1; i > 0; i--)
            {
                if (old_name[i] == '.')
                {
                    old_name[i] = ']';
                    old_name[i+1] = 0;
                    break;
                }
            }
            sprintf(&old_name[strlen(old_name)],"%s%s",init_node->name,
                    ".dir;1");
 
            /*
            **  find parent_node full spec
            */
            strcpy(new_name,parent_node->dir_name);
            sprintf(new_name,"%s%s%s",parent_node->dir_name,init_node->name,
                    ".dir;1");

            i = lib_rename(old_name,new_name,0);
            if (i != SS$_NORMAL)
            {
                sprintf(junk,"%s - failed to move %s%s%s",
                        full_error_text(i),"[.",init_node->name,"]");
                warning(junk);
            }
            else
            {
                /*
                ** make sure both structures get updated
                */
    	        curr_l_ptr->dp->to_be_saved = TRUE;
                init_line_ptr->dp->to_be_saved = TRUE;

                /*
                **  give init_node the old heave-hoh for a moment
                */
                delete_node(init_line_ptr, init_level);
    
                /*
                **  and add into the structure as the child of
                */
                add_node(curr_l_ptr, current_level, init_node);

                /*
                **  node has now a new did
                */
                init_node->did[0] = parent_node->fid[0];
                init_node->did[1] = parent_node->fid[1];
                init_node->did[2] = parent_node->fid[2];
            }
            /*
            **  rebuild lines
            */
            load_lines();
 
            /*
            **  clear message line
            */
            message("");
        }
        else
            warning("aborted");

        /*
        **  now find new line, level etc. for renamed node
        **  we might not have moved it
        */
        curr_l_ptr = first_line_ptr;
        for (current_line = 0; current_line < total_lines; current_line++)
        {
            for (current_level = 0; current_level < MAX_LEVEL; 
                 current_level++)
            {
                if (curr_l_ptr->n_ptr[current_level] == init_node)
                {
                    return(repair_display());
                }
            }
            curr_l_ptr = curr_l_ptr->forw_ptr;
        }

	/*
	**  not found - put cursor at top
	*/
        curr_l_ptr = first_line_ptr;
	current_line = 0;
	current_level = 0;
	repair_display();	
    } 
    else
        warning("Cannot move main directory");
}

purge_dirs()
{
    int status;

     show_deleting(curr_l_ptr->n_ptr[current_level],
                   DELETING_RENDITION,FALSE);
     if (confirm_user("Purge these directories Y/N [N] ?","Y"))
     {
         if (confirm_user("LAST CHANCE - Really purge these directories Y/N [N] ?","Y"))
         {
             delete_many_files(curr_l_ptr->n_ptr[current_level],TRUE,PURGE,"*.*;*");
             message("purged");
       }
    }
    show_deleting(curr_l_ptr->n_ptr[current_level],
                  NORMAL_RENDITION,FALSE);
    render(current_line,current_level,SPECIAL_RENDITION);
}
